# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2017, 2018. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# ******************************************************************

import numpy as np
import sys

from _snap_ml.utils import _is_mpi_enabled
_mpi_enabled = _is_mpi_enabled()

if sys.version_info[0] < 3:
    if _mpi_enabled:
        import libsnapmlmpi2 as libsnapml
    else:
        import libsnapmllocal2 as libsnapml
else:
    if _mpi_enabled:
        import _snap_ml.libsnapmlmpi3 as libsnapml
    else:
        import _snap_ml.libsnapmllocal3 as libsnapml

class DecisionTreeClassifier:

    """
    Decision Tree Classifier

    This class implements a decision tree classifier using the IBM Snap ML library.
    It can be used for binary classification problems. It handles both dense and 
    sparse matrix inputs. Use csr, csc or ndarray matrix format for training and csr or 
    ndarray format for prediction.

    Parameters
    ----------
    criterion : string, optional, default : "gini"
        This function measures the quality of a split. 
        Possible values: "gini" and "entropy" for information gain.
        "entropy" is currently not supported.

    splitter : string, optional, default : "best"
        This parameter defines the strategy used to choose the split at each node.
        Possible values: "best" and "random". "random" is currently not supported.

    max_depth : int or None, optional, default : None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than min_samples_leaf samples.

    min_samples_leaf : int or float, optional, default : 1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it generates at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  
        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then consider `ceil(min_samples_leaf * n_samples)` as the minimum number.
    
    max_features : int, float, string or None, optional, default : None
        The number of features to consider when looking for the best split:
            - If int, then consider `max_features` features at each split.
            - If float, then consider `int(max_features * n_features)` features at each split.
            - If "auto", then `max_features=sqrt(n_features)`.
            - If "sqrt", then `max_features=sqrt(n_features)`.
            - If "log2", then `max_features=log2(n_features)`.
            - If None, then `max_features=n_features`.

    random_state : int, or None, optional, default : None
        If int, random_state is the seed used by the random number generator;
        If None, the random number generator is the RandomState instance used by `np.random`.

    use_gpu : bool, default : False
        Flag that indicates the hardware platform used for training. If True, the training 
        is performed using the GPU. If False, the training is performed using the CPU. Currently
        only CPU training is supported.

    verbose : bool, default : False
        If True, it prints debugging information while training. Warning: this will increase the
        training time. For performance evaluation, use verbose=False.


    Attributes
    ----------
    classes_ : array of shape = [n_classes]
        The classes labels (single output problem)

    n_classes_ : int
        The number of classes (for single output problems)

    """

    def __init__(self,  criterion='gini', 
                        splitter='best', 
                        max_depth=None,
                        min_samples_leaf=1,
                        max_features=None,
                        random_state=None,
                        verbose=False, 
                        use_gpu=False):

        self.criterion_ = criterion
        if (not criterion == 'gini'): #and (not criterion == 'entropy'):
            raise ValueError("Parameter criterion: invalid value. Currently supported value is 'gini'.")	

        self.splitter_ = splitter
        if (not splitter == 'best'):
            raise ValueError("Parameter splitter: invalid value. Currently supported value is 'best'.")
        
        self.max_depth_ = max_depth
        if (not max_depth is None):
            if (not isinstance(max_depth, (int))):
                raise ValueError("Parameter max_depth: invalid value or type. Valid values are 'None' or any integer > 0.")
            elif (max_depth <= 0):
                raise ValueError("Parameter max_depth: invalid value. Valid values are 'None' or any integer > 0.")

        #self.min_samples_split_ = min_samples_split
        #if (not isinstance(min_samples_split, (int))) and (not isinstance(min_samples_split, (float))):
        #    raise TypeError("Parameter min_samples_split: invalid type. Valid type is int (any integer >= 2) or non-negative float in (0,1].")
        #elif (min_samples_split <= 0):
        #    raise ValueError("Parameter min_samples_split: invalid value. Valid values are any integer >= 2 or non-negative float in (0,1].")
        #elif (min_samples_split > 1) and (isinstance(min_samples_split, (float))):
        #    raise ValueError("Parameter min_samples_split: invalid value. Valid values are any integer >=2 or non-negative float in (0,1].")
        #elif (min_samples_split == 1) and (isinstance(min_samples_split, (int)):
        #    raise ValueError("Parameter min_samples_split: invalid value. Valid values are any integer >=2 or non-negative float in (0,1].")

        self.min_samples_leaf_ = min_samples_leaf
        if (not isinstance(min_samples_leaf, (int))) and (not isinstance(min_samples_leaf, (float))):
            raise TypeError("Parameter min_samples_leaf: invalid type. Valid type is int (any integer >= 1) or non-negative float in (0,1].")
        if (isinstance(min_samples_leaf, (int))) and (min_samples_leaf < 1):
            raise TypeError("Parameter min_samples_leaf: invalid value. Valid type is int (any integer >= 1) or non-negative float in (0,1].")
        if (isinstance(min_samples_leaf, (float))) and ((min_samples_leaf <= 0) or (min_samples_leaf > 1)):
            raise TypeError("Parameter min_samples_leaf: invalid value. Valid type is int (any integer >= 1) or non-negative float in (0,1].")

        self.max_features_ = max_features
        if (not max_features in [None, 'log2', 'sqrt', 'auto']):
            if (not isinstance(max_features, (int, float))):
                raise TypeError("Parameter max_features: invalid type. Valid type is int or float or string (log2, sqrt, auto) or None.")
            else:
                if (isinstance(max_features, (int))) and (max_features < 1):
                    raise ValueError("Parameter max_features: invalid value. Valid value is any integer >= 1 or float in (0,1].")
                if (isinstance(max_features, (float))) and ((max_features > 1) or (max_features <= 0)):
                    raise ValueError("Parameter max_features: invalid value. Valid value is any integer >=1 or float in (0,1].")

        # check on negative random_state values
        self.random_state_ = random_state
        if (not isinstance(random_state, (int))) and (not random_state is None):
            raise ValueType("Parameter random_state: invalid value or type. Valid type is int or None.")

        #self.max_leaf_nodes_ = max_leaf_nodes 
        #if (not max_leaf_nodes is None) and (not isinstance(max_leaf_nodes, (int))):
        #    raise ValueError("Parameter max_leaf_nodes: invalid value or type. Valid values are 'None' or any integer >= 1.")
        #if (isinstance(max_leaf_nodes, (int))) and (max_leaf_nodes < 1):
        #    raise ValueError("Parameter max_leaf_nodes: invalid value. Valid values are 'None' or any integer >=1.")

        #self.presort_ = presort
        #if (not isinstance(presort, (bool))) or (presort == True):
        #    raise TypeValue("Parameter presort: invalid type. Valid type is bool {True}.")
        
        self.verbose_ = verbose
        if (not isinstance(verbose, (bool))):
            raise TypeError("Parameter verbose: invalid type. Valid type is bool {True, False}.")

        #self.class_weight_ = class_weight
        #if (not class_weight == None):
        #    raise ValueError("Parameter class_weight: invalid parameter value. Valid values are {'None'}.")

        self.use_gpu_ = use_gpu
        if (not isinstance(use_gpu, (bool))):
            raise TypeError("Parameter use_gpu: invalid parameter type. Valid type is bool {False}.")
        elif (use_gpu == True):
            raise ValueError("Parameter use_gpu: invalid parameter value. Valid value is {False}. GPU acceleration not yet supported.")

        self.num_threads_ = 1
        #if (num_threads > 1):
        #    # currently multi-threading is not supported during training
        #    self.num_threads_ = 1 #num_threads
        #else:
        #    print("[Info] Training will run in single-threaded mode on CPU.")
        #    self.num_threads_ = 1

        if (_mpi_enabled):
            raise ValueError("MPI (distributed) implementation of decision tree classifiers not yet supported.")
 
    def get_params(self):

        """
        Get the values of the model parameters.

        Returns
        -------
        params : dict
        """

        params = {}

        params['criterion'] = self.criterion_
        params['splitter']  = self.splitter_
        params['max_depth'] = self.max_depth_
        #params['min_samples_split'] = self.min_samples_split_
        params['min_samples_leaf'] = self.min_samples_leaf_
        params['max_features'] = self.max_features_
        params['random_state'] = self.random_state_
        #params['max_leaf_nodes'] = self.max_leaf_nodes_
        #params['presort'] = self.presort_ 
        params['use_gpu'] = self.use_gpu_
        #params['class_weight'] = self.class_weight_
        params['verbose'] = self.verbose_
        #params['num_threads'] = self.num_threads_
        #params['min_impurity_decrease'] = self.min_impurity_decrease_

        return params

    def fit(self, X_train, y_train):

        """
        Fit the model according to the given train data.

        Parameters
        ----------
        X_train : sparse matrix (csr_matrix, csc_matrix) or dense matrix (ndarray)
            Train dataset

        y_train : array-like, shape = (n_samples,)
            The target vector corresponding to X_train. 

        Returns
        -------
        None
        """

        # Check if the balanced parameter was set
        #if (self.class_weight_ == 'balanced'):
        #    balanced = True
        #else:
        #    balanced = False

        # Check if the ginig criterion is used for evaluating the quality of a split
        if (self.criterion_ == 'gini'):
            gini_split_metric = True
        else:
            gini_split_metric = False

        # Check if the splitter is best or random
        if (self.splitter_ == 'best'):
            best_splitter = True
        else:
            best_splitter = False

        # Check if the maximum depth is limited by the user
        if (self.max_depth_ is None):
            max_depth = 0    # Unlimited tree depth
        else:
            max_depth = self.max_depth_

        # Random state
        if (self.random_state_ == None):
            # Not sure if this should be the random state
            random_state = np.random.get_state()[1][0]
        else:
            random_state = self.random_state_

        #if (self.max_leaf_nodes_ is None):
        #    max_leaf_nodes = 0    # No limitation on the number of leaf_nodes
        #else:
        #    max_leaf_nodes = self.max_leaf_nodes_

        num_ex = 0
        num_ft = 0
        num_nz = 0
        indptr = np.array([])
        indices = np.array([])
        data = np.array([])
        labs = np.array([])
        self.num_train_unique_labs = 0

        num_ex = len(y_train) 
        
        if (type(X_train).__name__ == "csr_matrix"):
            num_ft = X_train.shape[1]
        elif (type(X_train).__name__ == "csc_matrix"):
            X_train = X_train.tocsr()
            num_ft = X_train.shape[1]
        elif (type(X_train).__name__ == "ndarray"):
            if (X_train.shape[0] == X_train.shape[1]):
                print("[Warning] Each row in the input train dataset is assumed to be a training sample.")
                num_ft = X_train.shape[0]
            elif (X_train.shape[0] == num_ex):
                num_ft = X_train.shape[1]
            else:
                X_train = X_train.transpose().copy()
                num_ft = X_train.shape[1]

        if (type(X_train).__name__ == "csr_matrix" or type(X_train).__name__ == "csc_matrix"):
            num_nz    = X_train.nnz
            indptr    = X_train.indptr.astype(np.uint64)
            data      = X_train.data.astype(np.float32)
            indices   = X_train.indices.astype(np.uint32)
        elif (type(X_train).__name__ == "ndarray"):
            num_nz    = num_ex * num_ft
            data      = X_train.astype(np.float32)
            indptr    = np.array([])
            indices   = np.array([])
        else:
            raise TypeError("X_train should be in csr_matrix, csc_matrix or ndarray format.")
    
        # Extract set of unique labels from y_train
        self.uniq_labs_ = np.unique(y_train)
        self.classes_ = self.uniq_labs_
        self.n_classes_ = len(self.uniq_labs_)

        # Check the labels of y_train and train model(s)
        if (len(self.uniq_labs_) == 1):
            raise ValueError("There must be at least two unique label values in the train dataset.")
        elif (len(self.uniq_labs_) == 2):
            labs = y_train.astype(np.float32)
        else:
            raise ValueError("Multiclass classification not supported for decision tree classifiers.")

        # Check the values of min_samples_split
        #if (isinstance(self.min_samples_split_,(int))):
        #    if (self.min_samples_split_ > num_ex):
        #        raise ValueError("Parameter min_samples_split: invalid value. The value of min_samples_split is larger than the number of examples in the train dataset.")
        #    else:
        #        min_samples_split = self.min_samples_split_
        #else:
        #    min_samples_split = np.ceil(self.min_samples_split_ * num_ex)
        #    if (min_samples_split == 0):
        #        min_samples_split = 1 

        # Check the value of min_samples_leaf
        if (self.min_samples_leaf_ > num_ex):
            raise ValueError("Parameter min_samples_leaf: invalid value. The value of min_samples_leaf is larger than the number of examples in the train dataset.")
        else:
            min_samples_leaf = self.min_samples_leaf_

        # Check the value of max_features
        if (self.max_features_ is None):
            max_features = 0    # all features
        elif (self.max_features_ in ['auto', 'sqrt']):
            max_features = np.ceil(np.sqrt(num_ft)) 
        elif (self.max_features_ == 'log2'):
            import math
            max_features = np.ceil(math.log2(num_ft))
        elif (isinstance(self.max_features_, (int))):
            max_features = self.max_features_
        elif (isinstance(self.max_features_, (float))):
            max_features = np.ceil(self.max_features_*num_ft)
        else:
            raise ValueError("Parameter max_features: unable to parse.")

        max_features = int(max_features)
        if (max_features > num_ft):
            raise ValueError("Parameter max_features: invalid value. The value of max_features is larger than the number o features in the train dataset.")
                
        self.ptr_model_ = libsnapml.dtc_fit(gini_split_metric, 
                                    best_splitter, 
                                    max_depth,
                                    #min_samples_split,
                                    min_samples_leaf,
                                    #self.min_impurity_decrease_,
                                    max_features,
                                    random_state,
                                    #max_leaf_nodes,
                                    #self.presort_, 
                                    self.use_gpu_,
                                    #balanced,
                                    self.verbose_, 
                                    #self.num_threads_, 
                                    num_ex, 
                                    num_ft,
                                    num_nz, 
                                    indptr, 
                                    indices, 
                                    data, 
                                    labs)

        return None

    # The test dataset should be in csr_matrix format
    def predict_proba(self, X, num_threads=0):

        """
        Probability estimates

        The returned probability estimates for the two classes.
        Only for binary classification. 

        Parameters
        ----------
        X : sparse matrix (csr_matrix) or dense matrix (ndarray)
            Dataset used for predicting probability estimates.

        num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.

        Returns
        -------
        None
        """
        
        raise ValueError("predict_proba is currently not supported for decision tree classifiers. Use predict to compute the class estimates.")
            
        return None

    def predict(self, X, num_threads=0):

        """
        Class predictions

        The returned class estimates.

        Parameters
        ----------
        X : sparse matrix (csr_matrix) or dense matrix (ndarray)
            Dataset used for predicting class estimates.

        num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.

        Returns
        -------
        pred: array-like, shape = (n_samples,)
            Returns the predicted class of the sample.
        """

        num_ex = 0
        num_ft = 0
        indptr = np.array([])
        indices =  np.array([])
        data = np.array([])
        num_train_unique_labs = 0

        num_ex  = X.shape[0]
        num_ft  = X.shape[1]
        num_train_unique_labs = len(self.uniq_labs_)
        
        # libglm returns -1 and 1 as predicted classes
        # if the expected classes are 0 and 1, then we need to convert the predictions to 0 and 1
        # labels_flag keeps track of whether such a conversion is required
        labels_flag = False
        
        if (len(self.uniq_labs_) == 2) :
            if ((self.uniq_labs_[0] == 0) and (self.uniq_labs_[1] == 1)) or ((self.uniq_labs_[0] == 1) and (self.uniq_labs_[1] == 0)):
                labels_flag = True
        
        if (type(X).__name__ == "csc_matrix"):
            X = X.tocsr()
            indptr    = X.indptr.astype(np.uint64)
            data      = X.data.astype(np.float32)
            indices   = X.indices.astype(np.uint32)
        elif (type(X).__name__ == "csr_matrix"):
            indptr    = X.indptr.astype(np.uint64)
            data      = X.data.astype(np.float32)
            indices   = X.indices.astype(np.uint32)
        elif (type(X).__name__ == "ndarray"):
            data      = X.astype(np.float32)
            indptr    = np.array([])
            indices   = np.array([])
        else:
            raise ValueError("X should be in csc_matrix, csr_matrix or ndarray format.")

        pred = []

        # Generate predictions
        pred  = libsnapml.dtc_predict(num_ex,
                                      num_ft,
                                      indptr,
                                      indices,
                                      data,
                                      self.ptr_model_,
                                      num_threads)

        pred = np.array(pred)
        
        # convert labels from -1 1 to 0 1
        if (labels_flag == True):
            pred = (pred + 1)/2
        
        return pred

    # The test dataset should be in csr_matrix format
    def predict_log_proba(self, X, num_threads=0):

        """
        Log of probability estimates

        The returned log-probability estimates for the two classes.
        Only for binary classification.

        Parameters
        ----------
        X : sparse matrix (csr_matrix) or dense matrix (ndarray)
            Dataset used for predicting log-probability estimates.

        num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.

        Returns
        -------
        None
        """

        raise ValueError("predict_log_proba is currently not supported for decision tree classifiers. Use predict to compute the class estimates.")

        return None

