# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from pai4sk.cluster.dbscan_ import DBSCANSklearn
from ...exceptions import NotFittedError
import warnings
import numbers
import numpy as np
from .utils import _can_allow_cuml, _is_cuml_algorithm, _do_not_allow_cuml

class DBSCAN(DBSCANSklearn):
    """
    Perform DBSCAN clustering from vector array or distance matrix.

    DBSCAN - Density-Based Spatial Clustering of Applications with Noise. Finds
    core samples of high density and expands clusters from them. Good for data
    which contains clusters of similar density.

    If cuml is installed and if possible, then accelerated DBSCAN algorithm
    from cuML will be used. Otherwise, scikit-learn DBSCAN algorithm will be used.

    Parameters:
    ----------
    eps : float, optional
        The maximum distance between two samples for them to be considered as
        in the same neighborhood.

    min_samples : int, optional
        The number of samples (or total weight) in a neighborhood for a point
        to be considered as a core point. This includes the point itself.

    metric : string, or callable
        The metric to use when calculating distance between instances in a
        feature array. If metric is a string or callable, it must be one of
        the options allowed by pai4sk.metrics.pairwise_distances for its
        metric parameter. If metric is 'precomputed', X is assumed to be a
        distance matrix and must be square. X may be a sparse matrix, in which
        case only nonzero elements may be considered neighbors for DBSCAN.
        New in version 0.17: metric precomputed to accept precomputed sparse matrix.

    metric_params : dict, optional
        Additional keyword arguments for the metric function.
        New in version 0.19.

    algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute', 'cuml'}, optional
        The algorithm to be used by the NearestNeighbors module to compute
        pointwise distances and find nearest neighbors. See NearestNeighbors
        module documentation for details.

        If cuml is installed and if either (1) algorithm is set to "cuml" or
        (2) algorithm is "auto", then pai4sk will try to use DBSCAN algorithm
        from RAPIDS cuML if possible. cuML in pai4sk is currently supported
        only (a) with python 3.6 and (b) without MPI.

    leaf_size : int, optional (default = 30)
        Leaf size passed to BallTree or cKDTree. This can affect the speed of
        the construction and query, as well as the memory required to store the
        tree. The optimal value depends on the nature of the problem.

    p : float, optional
        The power of the Minkowski metric to be used to calculate distance between points.

    n_jobs : int or None, optional (default=None)
        The number of parallel jobs to run. None means 1 unless in a
        joblib.parallel_backend context. -1 means using all processors.
        See Glossary for more details.

    use_gpu : boolean, Default is True
        If True, cuML will use GPU 0. Applicable only for cuML.

    Attributes:
    core_sample_indices_ : array, shape = [n_core_samples]
        Indices of core samples.
    components_ : array, shape = [n_core_samples, n_features]
        Copy of each core sample found by training.
    labels_ : array, shape = [n_samples]
        Cluster labels for each point in the dataset given to fit(). Noisy
        samples are given the label -1.
    """

    __slots__ = ('_DBSCANSklearn')

    def __init__(self, eps=0.5, min_samples=5, metric='euclidean', metric_params=None, algorithm='auto', leaf_size=30, p=None, n_jobs=None, use_gpu=True):

        self.eps = eps
        self.min_samples = min_samples
        self.metric = metric
        self.metric_params = metric_params
        self.leaf_size = leaf_size
        self.p = p
        self.n_jobs = n_jobs
        self.algorithm = algorithm
        self.use_gpu = use_gpu
        self._cumlDBSCAN = None
        self.components_ = None
        self.labels_ = None
        self._DBSCANSklearn = super(DBSCAN, self)
        # Perform basic validations for enabling cuML first
        self._cuml_basic_validations = _can_allow_cuml()

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._DBSCANSklearn = super(DBSCAN, self)

    def fit(self, X, y=None, sample_weight=None):
        """Perform DBSCAN clustering from features or distance matrix.
        Parameters:
        ----------
        X : array or sparse (CSR) matrix of shape (n_samples, n_features), or array of shape (n_samples, n_samples)
            cuDF dataframe if cuml is installed.
            A feature array, or array of distances between samples if metric='precomputed'.
        sample_weight : array, shape (n_samples,), optional
            Weight of each sample, such that a sample with a weight of at least min_samples is by itself a core sample; a sample with negative weight may inhibit its eps-neighbor from being core. Note that weights are absolute, and default to 1.
        y : Ignored

        Returns:
        -------
        self : object
        """

        if self._check_set_cuml(X, sample_weight):
            print("Info: Invoking cuML dbscan fit...")
            self._cumlDBSCAN.fit(X)
            self.labels_ = self._cumlDBSCAN.labels_.to_array()
        else:
            _do_not_allow_cuml(X)
            print("Info: Invoking scikit-learn dbscan fit...")
            self._DBSCANSklearn.fit(X, y, sample_weight)

        return self

    def fit_predict(self, X, y=None, sample_weight=None):
        """
        Performs clustering on X and returns cluster labels.

        Parameters:
        X : array or sparse (CSR) matrix of shape (n_samples, n_features), or array of shape (n_samples, n_samples)
            A feature array, or array of distances between samples if metric='precomputed'.
        sample_weight : array, shape (n_samples,), optional
            Weight of each sample, such that a sample with a weight of at least min_samples is by itself a core sample; a sample with negative weight may inhibit its eps-neighbor from being core. Note that weights are absolute, and default to 1.

        y : Ignored

        Returns:
        -------
        y : ndarray, shape (n_samples,)
        """
        self.fit(X, y, sample_weight)
        return self.labels_

    def _check_set_cuml(self, X, sample_weight=None):
        """
           Check if cuML can be used for this DBSCAN
        """

        # Perform basic validations for enabling cuML first
        if not self._cuml_basic_validations:
            return False

        # Perform further vaildations for enabling cuML
        if not self.use_gpu:
            return False

        if not _is_cuml_algorithm(self.algorithm):
            return False

        if sample_weight is not None:
            warnings.warn("Currently cuML in pai4sk can't handle " \
                          "sample_weight argument.", UserWarning)
            return False

        if self.metric != 'euclidean':
            warnings.warn("Currently cuML in pai4sk can't handle " \
                          "metric argument.", UserWarning)
            return False

        if self.metric_params is not None:
            warnings.warn("Currently cuML in pai4sk can't handle " \
                          "metric_params argument.", UserWarning)
            return False

        if self.leaf_size != 30:
            warnings.warn("Currently cuML in pai4sk can't handle " \
                          "leaf_size argument.", UserWarning)
            return False

        if self.p is not None:
            warnings.warn("Currently cuML in pai4sk can't handle " \
                          "p argument.", UserWarning)
            return False

        if self.n_jobs is not None:
            warnings.warn("Currently cuML in pai4sk can't handle " \
                          "n_jobs argument.", UserWarning)
            return False

        import cudf
        if not isinstance(X, cudf.DataFrame) and not isinstance(X, np.ndarray):
            warnings.warn("Currently cuML in pai4sk can work only on " \
                          "cudf.DataFrame or numpy.ndarray.", UserWarning)
            return False

        if isinstance(X, np.ndarray):
            if X.dtype != np.float32 and X.dtype != np.float64:
                warnings.warn("Currently cuML in pai4sk can work on data of " \
                              "type numpy.float32 or numpy.float64.", UserWarning)
                return False

        if not self._cumlDBSCAN:
            from pai4sk import sml_utils
            ndevices = sml_utils.get_gpu_count()
            if ndevices == 0:
                warnings.warn("Currently cuML in pai4sk can work only with GPUs." \
                              "No GPUs are available.", UserWarning)
                return False

            from cuml import DBSCAN as cumldbscan
            self._cumlDBSCAN = cumldbscan(self.eps, self.min_samples)
            warnings.warn("Using DBSCAN from cuML.", UserWarning)

        return True

