# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019 . All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from pai4sk.decomposition.pca import PCASklearn
import warnings
import numpy as np
from .utils import _can_allow_cuml, _is_cuml_algorithm, _do_not_allow_cuml

class PCA(PCASklearn):
    """Principal component analysis (PCA)

    Linear dimensionality reduction using Singular Value Decomposition of the
    data to project it to a lower dimensional space.

    It uses the LAPACK implementation of the full SVD or a randomized truncated
    SVD by the method of Halko et al. 2009, depending on the shape of the input
    data and the number of components to extract.

    It can also use the scipy.sparse.linalg ARPACK implementation of the
    truncated SVD.

    Notice that this class does not support sparse input. See
    :class:`TruncatedSVD` for an alternative with sparse data.

    Read more in the :ref:`User Guide <PCA>`.

    Parameters
    ----------
    n_components : int, float, None or string
        Number of components to keep.
        if n_components is not set all components are kept::

            n_components == min(n_samples, n_features)

        If ``n_components == 'mle'`` and ``svd_solver == 'full'``, Minka's
        MLE is used to guess the dimension. Use of ``n_components == 'mle'``
        will interpret ``svd_solver == 'auto'`` as ``svd_solver == 'full'``.

        If ``0 < n_components < 1`` and ``svd_solver == 'full'``, select the
        number of components such that the amount of variance that needs to be
        explained is greater than the percentage specified by n_components.

        If ``svd_solver == 'arpack'``, the number of components must be
        strictly less than the minimum of n_features and n_samples.

        Hence, the None case results in::

            n_components == min(n_samples, n_features) - 1

    copy : bool (default True)
        If False, data passed to fit are overwritten and running
        fit(X).transform(X) will not yield the expected results,
        use fit_transform(X) instead.

    whiten : bool, optional (default False)
        When True (False by default) the `components_` vectors are multiplied
        by the square root of n_samples and then divided by the singular values
        to ensure uncorrelated outputs with unit component-wise variances.

        Whitening will remove some information from the transformed signal
        (the relative variance scales of the components) but can sometime
        improve the predictive accuracy of the downstream estimators by
        making their data respect some hard-wired assumptions.

    svd_solver : string {'auto', 'full', 'arpack', 'randomized', 'cuml', 'jacobi'}
        auto :
            when cuml is not used, the solver is selected by a default policy based
            on `X.shape` and `n_components`: if the input data is larger than 500x500 and the
            number of components to extract is lower than 80% of the smallest
            dimension of the data, then the more efficient 'randomized'
            method is enabled. Otherwise the exact full SVD is computed and
            optionally truncated afterwards. If cuml is used, then the default
            algorithm 'full' will be used when the svd_solver is 'auto' or 'cuml'.
        full :
            run exact full SVD calling the standard LAPACK solver via
            `scipy.linalg.svd` and select the components by postprocessing
        arpack :
            run SVD truncated to n_components calling ARPACK solver via
            `scipy.sparse.linalg.svds`. It requires strictly
            0 < n_components < min(X.shape)
        randomized :
            run randomized SVD by the method of Halko et al.

        .. versionadded:: 0.18.0

    tol : float >= 0, optional (default .0)
        Tolerance for singular values computed by svd_solver == 'arpack'.

        .. versionadded:: 0.18.0

    iterated_power : int >= 0, or 'auto', (default 'auto')
        Number of iterations for the power method computed by
        svd_solver == 'randomized'.
        Note : cuML for pai4sk only supports integer values for this parameter.

        .. versionadded:: 0.18.0

    random_state : int, RandomState instance or None, optional (default None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`. Used when ``svd_solver`` == 'arpack' or 'randomized'.

        .. versionadded:: 0.18.0

    use_gpu : boolean, Default is True
        If True, cuML will use GPU 0. Applicable only for cuML.

    Attributes
    ----------
    components_ : array, shape (n_components, n_features)
        Principal axes in feature space, representing the directions of
        maximum variance in the data. The components are sorted by
        ``explained_variance_``.

    explained_variance_ : array, shape (n_components,)
        The amount of variance explained by each of the selected components.

        Equal to n_components largest eigenvalues
        of the covariance matrix of X.

        .. versionadded:: 0.18

    explained_variance_ratio_ : array, shape (n_components,)
        Percentage of variance explained by each of the selected components.

        If ``n_components`` is not set then all components are stored and the
        sum of the ratios is equal to 1.0.

    singular_values_ : array, shape (n_components,)
        The singular values corresponding to each of the selected components.
        The singular values are equal to the 2-norms of the ``n_components``
        variables in the lower-dimensional space.

    mean_ : array, shape (n_features,)
        Per-feature empirical mean, estimated from the training set.

        Equal to `X.mean(axis=0)`.

    n_components_ : int
        The estimated number of components. When n_components is set
        to 'mle' or a number between 0 and 1 (with svd_solver == 'full') this
        number is estimated from input data. Otherwise it equals the parameter
        n_components, or the lesser value of n_features and n_samples
        if n_components is None.

    noise_variance_ : float
        The estimated noise covariance following the Probabilistic PCA model
        from Tipping and Bishop 1999. See "Pattern Recognition and
        Machine Learning" by C. Bishop, 12.2.1 p. 574 or
        http://www.miketipping.com/papers/met-mppca.pdf. It is required to
        compute the estimated data covariance and score samples.

        Equal to the average of (min(n_features, n_samples) - n_components)
        smallest eigenvalues of the covariance matrix of X.

    References
    ----------
    For n_components == 'mle', this class uses the method of `Minka, T. P.
    "Automatic choice of dimensionality for PCA". In NIPS, pp. 598-604`

    Implements the probabilistic PCA model from:
    `Tipping, M. E., and Bishop, C. M. (1999). "Probabilistic principal
    component analysis". Journal of the Royal Statistical Society:
    Series B (Statistical Methodology), 61(3), 611-622.
    via the score and score_samples methods.
    See http://www.miketipping.com/papers/met-mppca.pdf

    For svd_solver == 'arpack', refer to `scipy.sparse.linalg.svds`.

    For svd_solver == 'randomized', see:
    `Halko, N., Martinsson, P. G., and Tropp, J. A. (2011).
    "Finding structure with randomness: Probabilistic algorithms for
    constructing approximate matrix decompositions".
    SIAM review, 53(2), 217-288.` and also
    `Martinsson, P. G., Rokhlin, V., and Tygert, M. (2011).
    "A randomized algorithm for the decomposition of matrices".
    Applied and Computational Harmonic Analysis, 30(1), 47-68.`


    Examples
    --------
    >>> import numpy as np
    >>> from pai4sk.decomposition import PCA
    >>> X = np.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])
    >>> pca = PCA(n_components=2)
    >>> pca.fit(X)
    PCA(copy=True, iterated_power='auto', n_components=2, random_state=None,
      svd_solver='auto', tol=0.0, whiten=False)
    >>> print(pca.explained_variance_ratio_)  # doctest: +ELLIPSIS
    [0.9924... 0.0075...]
    >>> print(pca.singular_values_)  # doctest: +ELLIPSIS
    [6.30061... 0.54980...]

    >>> pca = PCA(n_components=2, svd_solver='full')
    >>> pca.fit(X)                 # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    PCA(copy=True, iterated_power='auto', n_components=2, random_state=None,
      svd_solver='full', tol=0.0, whiten=False)
    >>> print(pca.explained_variance_ratio_)  # doctest: +ELLIPSIS
    [0.9924... 0.00755...]
    >>> print(pca.singular_values_)  # doctest: +ELLIPSIS
    [6.30061... 0.54980...]

    >>> pca = PCA(n_components=1, svd_solver='arpack')
    >>> pca.fit(X)
    PCA(copy=True, iterated_power='auto', n_components=1, random_state=None,
      svd_solver='arpack', tol=0.0, whiten=False)
    >>> print(pca.explained_variance_ratio_)  # doctest: +ELLIPSIS
    [0.99244...]
    >>> print(pca.singular_values_)  # doctest: +ELLIPSIS
    [6.30061...]

    See also
    --------
    KernelPCA
    SparsePCA
    TruncatedSVD
    IncrementalPCA
    """

    __slots__ = ('_PCASklearn')

    def __init__(self, n_components=None, copy=True, whiten=False,
                 svd_solver='auto', tol=0.0, iterated_power='auto',
                 random_state=None, use_gpu=True):
        self.n_components = n_components
        self.copy = copy
        self.whiten = whiten
        self.svd_solver = svd_solver
        self.tol = tol
        self.iterated_power = iterated_power
        self.random_state = random_state
        self.use_gpu = use_gpu
        self.n_cols = None
        self.n_rows = None
        self._PCASklearn = super(PCA, self)
        self._cumlPCA = None
        # Perform basic validations for enabling cuML first
        self._cuml_basic_validations = _can_allow_cuml()

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._PCASklearn = super(PCA, self)

    def fit(self, X, y=None, _transform=True):
        """Fit the model with X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        y : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        if self._check_set_cuml(X):
            print("Info: Invoking cuML PCA fit...")
            self._cumlPCA.fit(X, _transform)
            self._save_fit_results()
        else:
            _do_not_allow_cuml(X, self.svd_solver, ['jacobi'])
            print("Info: Invoking scikit-learn PCA fit...")
            self._PCASklearn.fit(X, y)

        return self

    def fit_transform(self, X, y=None):
        """Fit the model with X and apply the dimensionality reduction on X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        y : Ignored

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)

        """
        if self._check_set_cuml(X):
            print("Info: Invoking cuML PCA fit_transform...")
            X_new = self._cumlPCA.fit_transform(X)
            self._save_fit_results()
            import cudf
            if isinstance(X_new, cudf.DataFrame):
                X_new = X_new.as_gpu_matrix(order='C').copy_to_host()
            return X_new
        else:
            _do_not_allow_cuml(X, self.svd_solver, ['jacobi'])
            print("Info: Invoking scikit-learn PCA fit_transform...")
            return self._PCASklearn.fit_transform(X, y)

    def transform(self, X):
        """Apply dimensionality reduction to X.

        X is projected on the first principal components previously extracted
        from a training set.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            New data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)

        Examples
        --------

        >>> import numpy as np
        >>> from pai4sk.decomposition import IncrementalPCA
        >>> X = np.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])
        >>> ipca = IncrementalPCA(n_components=2, batch_size=3)
        >>> ipca.fit(X)
        IncrementalPCA(batch_size=3, copy=True, n_components=2, whiten=False)
        >>> ipca.transform(X) # doctest: +SKIP
        """
        if self._cumlPCA:
            if not ((X.shape[0] > 1) and (X.shape[1] > 1)):
                raise ValueError("pai4sk cuML does not support n_rows and n_cols less than 2.")
            X_new = self._cumlPCA.transform(X)
            import cudf
            if isinstance(X_new, cudf.DataFrame):
                X_new = X_new.as_gpu_matrix(order='C').copy_to_host()
            return X_new
        else:
            print("Info: Invoking scikit-learn PCA transform...")
            return self._PCASklearn.transform(X)

    def inverse_transform(self, X):
        """Transform data back to its original space.

        In other words, return an input X_original whose transform would be X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_components)
            New data, where n_samples is the number of samples
            and n_components is the number of components.

        Returns
        -------
        X_original array-like, shape (n_samples, n_features)

        Notes
        -----
        If whitening is enabled, inverse_transform will compute the
        exact inverse operation, which includes reversing whitening.
        """
        if self._cumlPCA:
            print("Info: Invoking cuML PCA inverse_transform...")
            return self._cumlPCA.inverse_transform(X).as_gpu_matrix(order='C').copy_to_host()
        else:
            return self._PCASklearn.inverse_transform(X)

    def score(self, X, y=None):
        """Return the average log-likelihood of all samples.

        See. "Pattern Recognition and Machine Learning"
        by C. Bishop, 12.2.1 p. 574
        or http://www.miketipping.com/papers/met-mppca.pdf

        Parameters
        ----------
        X : array, shape(n_samples, n_features)
            The data.

        y : Ignored

        Returns
        -------
        ll : float
            Average log-likelihood of the samples under the current model
        """
        if self._cumlPCA:
            raise ValueError("Currently cuML in pai4sk does not support score method.")
        else:
            return self._PCASklearn.score(X)

    def _check_set_cuml(self, X):
        """
           Check if cuML can be used for this PCA
        """

        # Perform basic validations for enabling cuML first
        if not self._cuml_basic_validations:
            return False

        # Perform further vaildations for enabling cuML
        if not self.use_gpu:
            return False

        if not _is_cuml_algorithm(self.svd_solver, ['full', 'randomized', 'jacobi']):
            return False

        import cudf
        if not isinstance(X, cudf.DataFrame) and not isinstance(X, np.ndarray):
            warnings.warn("Currently cuML in pai4sk can work only on " \
                          "cudf.DataFrame or numpy.ndarray.", UserWarning)
            return False

        if isinstance(X, np.ndarray):
            if X.dtype != np.float32 and X.dtype != np.float64:
                warnings.warn("Currently cuML in pai4sk can work on data of " \
                              "type numpy.float32 or numpy.float64.", UserWarning)
                return False

        if (self.n_components is not None):
            if ((type(self.n_components) is not int) or (self.n_components <= 0)):
                warnings.warn("Currently cuML in pai4sk can work with n_components " \
                              "parameter of only positive integer value.", UserWarning)
                return False

        if (self.iterated_power is not "auto") and (type(self.iterated_power) is not int):
            warnings.warn("Currently cuML in pai4sk can only work with integer" \
                          "values for parameter 'iterated_power'.", UserWarning)
            return False

        if not self._cumlPCA:
            if (self.iterated_power is None or self.iterated_power is "auto"):
                self.iterated_power = 15
            if self.n_components is None:
                self.n_components = min(X.shape)

            from cuml import PCA as cumlPCA
            self._cumlPCA = cumlPCA(n_components=self.n_components, copy=self.copy, whiten=self.whiten,
                                    tol=self.tol, iterated_power=self.iterated_power,
                                    random_state=self.random_state, svd_solver=self.svd_solver)
            warnings.warn("Using PCA from cuML.", UserWarning)

        return True

    def _save_fit_results(self):
        """
           Save the results from cuML fit
        """
        df = self._cumlPCA.components_
        self.components_ = df.as_gpu_matrix(order='C').copy_to_host()
        explained_variance_cudf_series = self._cumlPCA.explained_variance_
        self.explained_variance_ = explained_variance_cudf_series.to_array()
        explained_variance_ratio_cudf_series = self._cumlPCA.explained_variance_ratio_
        self.explained_variance_ratio_ = explained_variance_ratio_cudf_series.to_array()
        singular_values_cudf_series = self._cumlPCA.singular_values_
        self.singular_values_ = singular_values_cudf_series.to_array()
        self.mean_ = self._cumlPCA.mean_
        self.n_components_ = self.n_components
        noise_variance_cudf_series = self._cumlPCA.noise_variance_
        self.noise_variance_ = noise_variance_cudf_series.to_array()

