# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from pai4sk.linear_model.ridge import RidgeSklearn
from ...utils.validation import check_X_y
from ...utils.multiclass import check_classification_targets
from ...utils.extmath import safe_sparse_dot
import warnings
import numpy as np
from scipy import sparse
from .utils import _check_snapml_types, _check_snapml_type, _is_gpu_matrix_type
from _snap_ml.utils import _is_mpi_enabled, _is_data_snapml_partition_type
_mpi_enabled = _is_mpi_enabled()

class Ridge(RidgeSklearn):
    """Linear least squares with l2 regularization.

    Minimizes the objective function::

    ||y - Xw||^2_2 + alpha * ||w||^2_2

    This model solves a regression model where the loss function is
    the linear least squares function and regularization is given by
    the l2-norm. Also known as Ridge Regression or Tikhonov regularization.
    This estimator has built-in support for multi-variate regression
    (i.e., when y is a 2d-array of shape [n_samples, n_targets]).

    Read more in the :ref:`User Guide <ridge_regression>`.

    For SnapML solver this supports both local and distributed(MPI) method of execution.

    Parameters
    ----------
    alpha : {float, array-like}, shape (n_targets)
        Regularization strength; must be a positive float. Regularization
        improves the conditioning of the problem and reduces the variance of
        the estimates. Larger values specify stronger regularization.
        Alpha corresponds to ``C^-1`` in other linear models such as
        LogisticRegression or LinearSVC. If an array is passed, penalties are
        assumed to be specific to the targets. Hence they must correspond in
        number.

    fit_intercept : boolean
        Whether to calculate the intercept for this model. If set
        to false, no intercept will be used in calculations
        (e.g. data is expected to be already centered).

    normalize : boolean, optional, default False
        This parameter is ignored when ``fit_intercept`` is set to False.
        If True, the regressors X will be normalized before regression by
        subtracting the mean and dividing by the l2-norm.
        If you wish to standardize, please use
        :class:`pai4sk.preprocessing.StandardScaler` before calling ``fit``
        on an estimator with ``normalize=False``.

    copy_X : boolean, optional, default True
        If True, X will be copied; else, it may be overwritten.

    max_iter : int, optional
        Maximum number of iterations for conjugate gradient solver.
        For 'sparse_cg' and 'lsqr' solvers, the default value is determined
        by scipy.sparse.linalg. For 'sag' solver, the default value is 1000.

    tol : float
        Precision of the solution.

    regularizer : float, default : 1.0
        Regularization strength. It must be a positive float.
        Larger regularization values imply stronger regularization.

    use_gpu : bool, default : False
        Flag for indicating the hardware platform used for training. If True, the training
        is performed using the GPU. If False, the training is performed using the CPU.
        Applicable only for snapml solver

    device_ids : array-like of int, default : []
        If use_gpu is True, it indicates the IDs of the GPUs used for training.
        For single GPU training, set device_ids to the GPU ID to be used for training, e.g., [0].
        For multi-GPU training, set device_ids to a list of GPU IDs to be used for training, e.g., [0, 1].
        Applicable only for snapml solver

    num_threads : int, default : 1
        The number of threads used for running the training. The value of this parameter
        should be a multiple of 32 if the training is performed on GPU (use_gpu=True)
        (default value for GPU is 256). Applicable only for snapml solver

    return_training_history : str or None, default : None
        How much information about the training should be collected and returned by the fit function. By 
        default no information is returned (None), but this parameter can be set to "summary", to obtain 
        summary statistics at the end of training, or "full" to obtain a complete set of statistics 
        for the entire training procedure. Note, enabling either option will result in slower training.
        Applicable only for snapml solver

    solver : {'auto', 'svd', 'cholesky', 'lsqr', 'sparse_cg', 'sag', 'saga', 'snapml'}
        Solver to use in the computational routines:

        - 'auto' chooses the solver automatically based on the type of data.

        - 'svd' uses a Singular Value Decomposition of X to compute the Ridge
          coefficients. More stable for singular matrices than
          'cholesky'.

        - 'cholesky' uses the standard scipy.linalg.solve function to
          obtain a closed-form solution.

        - 'sparse_cg' uses the conjugate gradient solver as found in
          scipy.sparse.linalg.cg. As an iterative algorithm, this solver is
          more appropriate than 'cholesky' for large-scale data
          (possibility to set `tol` and `max_iter`).

        - 'lsqr' uses the dedicated regularized least-squares routine
          scipy.sparse.linalg.lsqr. It is the fastest and uses an iterative
          procedure.

        - 'sag' uses a Stochastic Average Gradient descent, and 'saga' uses
          its improved, unbiased version named SAGA. Both methods also use an
          iterative procedure, and are often faster than other solvers when
          both n_samples and n_features are large. Note that 'sag' and
          'saga' fast convergence is only guaranteed on features with
          approximately the same scale. You can preprocess the data with a
          scaler from pai4sk.preprocessing.

        All last five solvers support both dense and sparse data. However,
        only 'sag' and 'saga' supports sparse input when `fit_intercept` is
        True.

        .. versionadded:: 0.17
           Stochastic Average Gradient descent solver.
        .. versionadded:: 0.19
           SAGA solver.

    random_state : int, RandomState instance or None, optional, default None
        The seed of the pseudo random number generator to use when shuffling
        the data.  If int, random_state is the seed used by the random number
        generator; If RandomState instance, random_state is the random number
        generator; If None, the random number generator is the RandomState
        instance used by `np.random`. Used when ``solver`` == 'sag'.

        .. versionadded:: 0.17
           *random_state* to support Stochastic Average Gradient.

    privacy : bool, default : False
        Train the model using a differentially private algorithm.

    eta : float, default : 0.3
        Learning rate for the differentially private training algorithm.

    batch_size : int, default : 100
        Mini-batch size for the differentially private training algorithm.

    privacy_epsilon : float, default : 10.0
        Target privacy gaurantee. Learned model will be (privacy_epsilon, 0.01)-private.

    grad_clip : float, default: 1.0
        Gradient clipping parameter for the differentially private training algorithm

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Weight vector(s).

    intercept_ : float | array, shape = (n_targets,)
        Independent term in decision function. Set to 0.0 if
        ``fit_intercept = False``.

    n_iter_ : array or None, shape (n_targets,)
        Actual number of iterations for each target. Available only for
        sag and lsqr solvers. Other solvers will return None.

    training_history_ : dict
        It returns a dictionary with the following keys : 'epochs', 't_elap_sec', 'train_obj'.
        If 'return_training_history' is set to "summary", 'epochs' contains the total number of
        epochs performed, 't_elap_sec' contains the total time for completing all of those epochs.
        If 'return_training_history' is set to "full", 'epochs' indicates the number of epochs
        that have elapsed so far, and 't_elap_sec' contains the time to do those epochs.
        'train_obj' is the training loss.
        Applicable only for snapml solver.

        .. versionadded:: 0.17

    See also
    --------
    RidgeClassifier : Ridge classifier
    RidgeCV : Ridge regression with built-in cross validation
    :class:`pai4sk.kernel_ridge.KernelRidge` : Kernel ridge regression
        combines ridge regression with the kernel trick

    Examples
    --------
    >>> from pai4sk.linear_model import Ridge
    >>> import numpy as np
    >>> n_samples, n_features = 10, 5
    >>> np.random.seed(0)
    >>> y = np.random.randn(n_samples)
    >>> X = np.random.randn(n_samples, n_features)
    >>> clf = Ridge(alpha=1.0)
    >>> clf.fit(X, y) # doctest: +NORMALIZE_WHITESPACE
    Ridge(alpha=1.0, copy_X=True, fit_intercept=True, max_iter=None,
          normalize=False, random_state=None, solver='auto', tol=0.001)

    """
    __slots__ = ('_ridgesklearn')

    def __init__(self, alpha=1.0, fit_intercept=True, normalize=False,
                 copy_X=True, max_iter=None, tol=1e-3, solver="auto",
                 random_state=None,
                 dual=False,                        # pai4sk
                 verbose=0,                         # pai4sk
                 use_gpu=True,                      # pai4sk
                 device_ids=[],                     # pai4sk
                 return_training_history=None,      # pai4sk
                 privacy=False,                     # pai4sk
                 eta=0.3,                           # pai4sk
                 batch_size=100,                    # pai4sk 
                 privacy_epsilon=10,                # pai4sk
                 grad_clip=1,                       # pai4sk
                 num_threads=1):                    # pai4sk
        self.alpha = alpha
        self.fit_intercept = fit_intercept
        self.normalize = normalize
        self.copy_X = copy_X
        self.max_iter = max_iter
        self.tol = tol
        self.solver = solver
        self.random_state = random_state
        self.verbose = verbose
        self.use_gpu = use_gpu
        self.dual = dual
        self.device_ids = device_ids
        self.num_threads = num_threads
        self.return_training_history = return_training_history
        super(Ridge, self).__init__(alpha=self.alpha, fit_intercept=self.fit_intercept, 
                normalize=self.normalize, copy_X=self.copy_X, max_iter=self.max_iter, 
                tol=self.tol, solver=self.solver, random_state=self.random_state)
        self._ridgesklearn = super(Ridge, self)
        self._snaplr = None
        self.privacy = privacy
        self.eta = eta
        self.batch_size = batch_size
        self.privacy_epsilon = privacy_epsilon
        self.grad_clip = grad_clip

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._ridgesklearn = super(Ridge, self).__init__(alpha=self.alpha, 
                fit_intercept=self.fit_intercept,
                normalize=self.normalize, copy_X=self.copy_X, max_iter=self.max_iter,
                tol=self.tol, solver=self.solver, random_state=self.random_state)

    def _check_set_snapml(self, X, y, sample_weight=None):
        if self.solver is not 'snapml' and self.solver is not 'auto':
            return False

        if not sample_weight is None:
            return False

        if not _check_snapml_types(X,y):
            if self._snaplr:
                raise ValueError("SnapML: unsupported data types")
            return False

        if not _is_data_snapml_partition_type(X) and not _is_gpu_matrix_type(X, y):
            # Handle small dataset
            if X.shape[0] < 100:
                if self.solver == 'snapml':
                    self.solver = 'auto'
                    warnings.warn("PowerAI: switch to 'auto' solver as the input size is less")
                return False

            # can't handle multi-variate
            if not y.ndim == 1:
                if self._snaplr:
                    raise ValueError("SnapML: unsupported label dimension")
                if self.solver == 'snapml':
                    self.solver = 'auto'
                return False

        if not self._snaplr:
            # Check for GPUs
            if self.use_gpu:
                from pai4sk import sml_utils
                ndevices = sml_utils.get_gpu_count()
                if ndevices > 0 and not self.device_ids:
                    if _is_gpu_matrix_type(X, y):
                        self.device_ids = list(range(0, 1)) #for GPU data frame always GPU-id 0 is selected
                    else:
                        self.device_ids = list(range(0, ndevices))
                elif ndevices == 0:
                    self.use_gpu = False

            from pai4sk import LinearRegression as _snaplr
            self._snaplr = _snaplr(use_gpu=False if self.use_gpu is None else self.use_gpu,
                                    device_ids=[] if self.device_ids is None else self.device_ids,
                                    max_iter=1000 if self.max_iter is None else self.max_iter,
                                    num_threads=self.num_threads,
                                    verbose=False if self.verbose == 0 else True,
                                    regularizer=1 if self.alpha is None else self.alpha,
                                    tol=0.001 if self.tol is None else self.tol,
                                    fit_intercept=False if self.fit_intercept is None else self.fit_intercept,
                                    intercept_scaling=1.0 if self.normalize is None else self.normalize,
                                    return_training_history=self.return_training_history,
                                    privacy=self.privacy, eta=self.eta, batch_size=self.batch_size,
                                    privacy_epsilon=self.privacy_epsilon, grad_clip=self.grad_clip,
                                    dual=self.dual, penalty="l2")

            if self.solver == 'auto' and self._snaplr:
                self.solver = 'snapml'
                warnings.warn("PowerAI: Default solver is set to 'snapml'. "
                          "Specify a solver to silence this warning.")
        else:
            if not self.solver == 'snapml':
                self.solver = 'snapml'
                warnings.warn("PowerAI: switch back to 'snapml' solver")

        return True

    def fit(self, X, y, sample_weight=None):
        """Fit Ridge regression model

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training data
            For SnapML solver it also supports input of types SnapML data partition and DeviceNDArray.

        y : array-like, shape = [n_samples] or [n_samples, n_targets]
            Target values

        sample_weight : float or numpy array of shape [n_samples]
            Individual weights for each sample

        Returns
        -------
        self : returns an instance of self.
        """
        if self.solver in ('sag', 'saga'):
            _dtype = np.float64
        else:
            # all other solvers work at both float precision levels
            _dtype = [np.float64, np.float32]

        if not _is_data_snapml_partition_type(X) and not _is_gpu_matrix_type(X, y):
            if not _mpi_enabled:
                X, y = check_X_y(X, y, ['csr', 'csc', 'coo'], dtype=_dtype,
                                 multi_output=True, y_numeric=True)

        if ((sample_weight is not None) and
                np.atleast_1d(sample_weight).ndim > 1):
            raise ValueError("Sample weights must be 1D array or scalar")

        if self._check_set_snapml(X, y, sample_weight):
            X, y, X_offset, y_offset, X_scale = self._preprocess_data(
                X, y, self.fit_intercept, self.normalize, self.copy_X,
                sample_weight=sample_weight)

            self.training_history_ = self._snaplr.fit(X, y)

            if not _mpi_enabled:
                # n_iter_ not returned for MPI
                self.n_iter_ = self._snaplr.n_iter_

            if (self.fit_intercept):
                self.coef_ = self._snaplr.coef_[:-1]
                if not sparse.issparse(X):
                    self._set_intercept(X_offset, y_offset, X_scale)
                else:
                    self.intercept_ = self._snaplr.coef_[-1]
            else:
                self.coef_ = self._snaplr.coef_
                self.intercept_ = 0.0
            return self
        else:
            rd_sk = self._ridgesklearn.fit(X, y, sample_weight)
            self.coef_ = rd_sk.coef_
            self.intercept_ = rd_sk.intercept_
            self.n_iter_ = rd_sk.n_iter_
            self.training_history_ = []
            return self

    def predict(self, X, num_threads=0):

        """
        Class predictions
        The returned class estimates.
        Parameters
        ----------
        X : sparse matrix (csr_matrix) or dense matrix (ndarray)
            Dataset used for predicting class estimates.
            For SnapML solver it also supports input of type SnapML data partition.
       num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.
        Returns
        -------
        proba: array-like, shape = (n_samples,)
            Returns the predicted class of the sample.
        """ 
        if not hasattr(self, "coef_"):
            raise NotFittedError("Call fit before prediction")
        
        if _check_snapml_type(X) and self.solver == 'snapml':
            if self.fit_intercept and not sparse.issparse(X):
                return safe_sparse_dot(X, self.coef_.T,
                                   dense_output=True) + self.intercept_
            else:
                return self._snaplr.predict(X, num_threads)
        else:
            return self._ridgesklearn.predict(X)


