# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

import time
import numpy as np
import sys
import argparse
import pandas as pd

defaultPath = "."

CLI=argparse.ArgumentParser()
CLI.add_argument(
   "--device_ids",
   nargs="*",
   type=int,
   default=[]
)
CLI.add_argument(
    "--skip_sklearn",
    action = 'store_true',
    default=False
)
CLI.add_argument(
    "--data_path",
    type=str,
    default=defaultPath
)
CLI.add_argument(
   "--use_balanced_class_weights",
   action = 'store_true',
   default=False
)
CLI.add_argument(
   "--fit_intercept",
   action = 'store_true',
   default=False
)

args = CLI.parse_args()
t0 = time.time()

# Load data
pdf_trainX = pd.read_csv(args.data_path + "/data/epsilon.X_train.csv", dtype=np.float32, header=None)
pdf_testX  = pd.read_csv(args.data_path + "/data/epsilon.X_test.csv", dtype=np.float32, header=None)
pdf_trainY = pd.read_csv(args.data_path + "/data/epsilon.y_train.csv", dtype=np.float32, header=None)
pdf_testY  = pd.read_csv(args.data_path + "/data/epsilon.y_test.csv", dtype=np.float32, header=None);

print("Data load time (s):  {0:.2f}".format(time.time()-t0))

# Convert pandas to cudf dataframes
import cudf
from cudf.dataframe import DataFrame

df_trainX = DataFrame.from_pandas(pdf_trainX)
df_trainY = DataFrame.from_pandas(pdf_trainY)

# ndarray data used for training (for sklearn case)
# Converting to C-contiguous ndarray
X_train_ndarray = np.ascontiguousarray(pdf_trainX.values)
y_train_ndarray = np.ascontiguousarray(pdf_trainY.values)

# data used for training
# Converting to C-contiguous DeviceNDArray
X_train = df_trainX.as_gpu_matrix(order='C')
y_train = df_trainY.as_gpu_matrix(order='C')

#data used for inferencing
# Converting to C-contiguous ndarray
X_test = np.ascontiguousarray(pdf_testX.values)
y_test = np.ascontiguousarray(pdf_testY.values)

num_threads = 256
cpu_gpu = "GPU"

if args.use_balanced_class_weights:
  class_weight = "balanced"
else:
  class_weight = None

# Create a SVM instance from pai4sk explicit call
from pai4sk import SupportVectorMachine
svm = SupportVectorMachine(use_gpu=True, num_threads=num_threads, 
                           class_weight=class_weight, device_ids=args.device_ids, 
                           regularizer=2, fit_intercept=args.fit_intercept)

# Training
t0 = time.time()
svm.fit(X_train, y_train)
print("[pai4sk] {0} Training time (s):  {1:.2f}".format("GPU", time.time()-t0))


# Inference
pred = svm.predict(X_test)

# Evaluate accuracy on test set
from pai4sk.metrics import accuracy_score
acc_snap = accuracy_score(y_test, pred)
print("[pai4sk] {0} Accuracy: {1:.3f}".format(cpu_gpu, acc_snap))

# Inference
pred = svm.decision_function(X_test)


#This is another way of import

# Create a SVM instance from pai4sk implicit call
from pai4sk.svm import LinearSVC as SupportVectorMachine 
svm = SupportVectorMachine(use_gpu=True, num_threads=num_threads,
                           class_weight=class_weight, device_ids=args.device_ids,
                           fit_intercept=args.fit_intercept, loss="hinge", C=0.5, dual=True, tol=0.001)

# Training
t0 = time.time()
svm.fit(X_train, y_train)
print("[pai4sk.svm] {0} Training time (s):  {1:.2f}".format("GPU", time.time()-t0))


# Inference
pred = svm.predict(X_test)


# Evaluate accuracy on test set
from pai4sk.metrics import accuracy_score
acc_snap = accuracy_score(y_test, pred)
print("[pai4sk.svm] {0} Accuracy: {1:.3f}".format(cpu_gpu, acc_snap))

# Inference
pred = svm.decision_function(X_test)

if args.skip_sklearn:
    print('Execution using scikit-learn is skipped.')
    exit()



# Create a SVM instance from pai4sk explicit call
from pai4sk import SupportVectorMachine
svm = SupportVectorMachine(use_gpu=True, num_threads=num_threads,
                           class_weight=class_weight, device_ids=args.device_ids,
                           regularizer=2, fit_intercept=args.fit_intercept)

# Training
t0 = time.time()
svm.fit(X_train_ndarray, y_train_ndarray)
print("[pai4sk] {0} Training time (s):  {1:.2f}".format("GPU", time.time()-t0))


# Inference
pred = svm.predict(X_test)

# Evaluate accuracy on test set
from pai4sk.metrics import accuracy_score
acc_snap = accuracy_score(y_test, pred)
print("[pai4sk] {0} Accuracy: {1:.3f}".format(cpu_gpu, acc_snap))

# Inference
pred = svm.decision_function(X_test)
