# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019 . All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

# Sample program to use PCA from cuML in paisk

import sys
if sys.version_info.major != 3:
    raise ValueError("cuML in pai4sk is supported only with python 3.6.")

import numpy as np
import pandas as pd
from pai4sk.decomposition import PCA as pai4skPCA
from sklearn.decomposition import PCA as skPCA
import cudf
import os
import time

import argparse

CLI=argparse.ArgumentParser()
CLI.add_argument(
   "--nrows",
   type=int,
   default=2**20
)
args = CLI.parse_args()

import gzip
def load_data(nrows, ncols, cached = 'data/mortgage.npy.gz'):
    if os.path.exists(cached):
        print('use mortgage data')
        with gzip.open(cached) as f:
            X = np.load(f)
        X = X[np.random.randint(0,X.shape[0]-1,nrows),:ncols]
    else:
        print('use random data')
        X = np.random.rand(nrows,ncols)
    df = pd.DataFrame({'fea%d'%i:X[:,i] for i in range(X.shape[1])})
    return df

from sklearn.metrics import mean_squared_error
def array_equal(a,b,threshold=2e-3,with_sign=True):
    a = to_nparray(a)
    b = to_nparray(b)
    if with_sign == False:
        a,b = np.abs(a),np.abs(b)
    error = mean_squared_error(a,b)
    res = error<threshold
    return res

def to_nparray(x):
    if isinstance(x,np.ndarray) or isinstance(x,pd.DataFrame):
        return np.array(x)
    elif isinstance(x,np.float64):
        return np.array([x])
    elif isinstance(x,cudf.DataFrame) or isinstance(x,cudf.Series):
        return x.to_pandas().values
    return x

nrows = args.nrows
ncols = 400

X = load_data(nrows,ncols)
print('data',X.shape)

n_components = 10
whiten = False
random_state = 42
svd_solver="auto"

pca_sk = skPCA(n_components=n_components,svd_solver=svd_solver, 
            whiten=whiten, random_state=random_state)
t0 = time.time()
result_sk = pca_sk.fit_transform(X)
print("[sklearn] Training time (s):  {0:.2f}".format(time.time()-t0))

X = cudf.DataFrame.from_pandas(X)

pca_pai4sk = pai4skPCA(n_components=n_components,svd_solver=svd_solver, 
            whiten=whiten, random_state=random_state)
t0 = time.time()
result_pai4sk = pca_pai4sk.fit_transform(X)
print("[pai4sk] Training time (s):  {0:.2f}".format(time.time()-t0))


for attr in ['singular_values_','components_','explained_variance_',
             'explained_variance_ratio_']:
    passed = array_equal(getattr(pca_sk,attr),getattr(pca_pai4sk,attr),threshold=0.5)
    message = 'compare PCA: pai4sk vs sklearn {:>25} {}'.format(attr,'equal' if passed else 'NOT equal')
    print(message)

passed = array_equal(result_sk,result_pai4sk,threshold=0.8)
message = 'compare PCA: pai4sk vs sklearn transformed results %s'%('equal'if passed else 'NOT equal')
print(message)


